/*
 *  ---------
 * |.**> <**.|  CardContact
 * |*       *|  Software & System Consulting
 * |*       *|  Minden, Germany
 * |'**> <**'|  Copyright (c) 2002. All rights reserved
 *  --------- 
 *
 * See file LICENSE for details on licensing
 *
 * Abstract :       Defines procedures for ifdhandler 2.0 for pcsc-lite 
 *
 * Author :         Frank Thater (FTH)
 *
 * Last modified:   12/06/2002
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pthread.h>
#include <errno.h>

#include "pcscdefines.h"
#include "ifd_handler.h"

#include "atr.h"
#include "defines.h"
#include "ctapi.h"
#include "sercom.h"
#include "eco5000.h"
#include "ctbcs.h"
#include "ecotools.h"


extern struct eco5000_t *ecoTable[MAX_READER];

static pthread_mutex_t ifdh_status_mutex = PTHREAD_MUTEX_INITIALIZER;


/*
 * Locate matching LUN in table of active readers. (IFD-handler)
 *
 */

static int LookupLun(unsigned long lun)

{
    int i;

    for (i = 0; i < MAX_READER; i++) {
        if (ecoTable[i] && (ecoTable[i]->lun == lun)) {
             return i;
        }
    }
    return -1;
}


/*
 * Get the ATR of a card and determine if the present card is an
 * async or sync card.
 * Fill the needed fields in the context structure of the IFD.
 *
 */

int IFDGetATR (struct eco5000_t *ctx)

{

    int response;

    memset(ctx->ATR, 0, sizeof(ctx->ATR));              /* Clear parts of the context */
    memset(ctx->HCC, 0, sizeof(ctx->HCC));              
   
    ctx->Protocol = 0;
    ctx->NumOfHB = 0;
    ctx->LenOfATR = 0;
   
    if ((response = ASync_GetATR(ctx)) == 1) {
        if((response = Sync_GetATR(ctx) < 0))
            return ERR_CT;;
        ctx->Type = SYNC_ICC;
    } else if(response < 0) {
        return response;
    } else {

        ctx->Type = ASYNC_ICC;

    }
    
    return 0;
    
}


RESPONSECODE IFDHCreateChannel ( DWORD Lun, DWORD Channel ) {

  /* Lun - Logical Unit Number, use this for multiple card slots 
     or multiple readers. 0xXXXXYYYY -  XXXX multiple readers,
     YYYY multiple slots. The resource manager will set these 
     automatically.  By default the resource manager loads a new
     instance of the driver so if your reader does not have more than
     one smartcard slot then ignore the Lun in all the functions.
     Future versions of PC/SC might support loading multiple readers
     through one instance of the driver in which XXXX would be important
     to implement if you want this.
  */
  
  /* Channel - Channel ID.  This is denoted by the following:
     0x000001 - /dev/pcsc/1
     0x000002 - /dev/pcsc/2
     0x000003 - /dev/pcsc/3
     
     USB readers may choose to ignore this parameter and query 
     the bus for the particular reader.
  */

  /* This function is required to open a communications channel to the 
     port listed by Channel.  For example, the first serial reader on COM1 would
     link to /dev/pcsc/1 which would be a sym link to /dev/ttyS0 on some machines
     This is used to help with intermachine independance.
     
     Once the channel is opened the reader must be in a state in which it is possible
     to query IFDHICCPresence() for card status.
 
     returns:

     IFD_SUCCESS
     IFD_COMMUNICATION_ERROR
  */

    int rc, retry, indx;
    HANDLE fh;
    char device[20];
    unsigned short pn;
    unsigned short highnibble;
    unsigned short lownibble;
    
    struct eco5000_t *ctx;

#ifdef DEBUG
    printf("Lun=%x Channel=%x\n", Lun, Channel);
#endif
    
    rc = LookupLun(Lun);

    if (rc >= 0) {
        return IFD_COMMUNICATION_ERROR;
    }
    
    lownibble = Channel >> 16;

    highnibble = Channel & 0xFFFF;
    
    if (lownibble == 0x01)  {
    	    // Serial Port.

            switch (highnibble)
            {

                case 0x3F8:
                    pn = 0;
                    break;

                case 0x2F8:
                    pn = 1;
                    break;

                case 0x3E8:
                    pn = 2;
                    break;

                case 0x2E8:
                    pn = 3;
                    break;

                default:
                    return IFD_NOT_SUPPORTED;
	    }
    } else {
        return IFD_NOT_SUPPORTED;
    }
    
    sprintf(device, "/dev/ttyS%d", pn);
    
#ifdef DEBUG
    printf("Opening reader on %s\n", device);
#endif

    rc = rs232Open(&fh, device, 9600, 'E', 8, 2, 200);

    if (rc < 0) {
        return IFD_COMMUNICATION_ERROR;        /* Could not allocate port */
    }
    
    ctx = (struct eco5000_t *)malloc(sizeof(struct eco5000_t));

    if (!ctx) {
        return IFD_COMMUNICATION_ERROR;
    }
    
    ctx->ctn = 0x00;
    ctx->lun = Lun;
    ctx->fh = fh;
    ctx->CTModFunc = NULL;
    ctx->RXStatus = OFF;
    ctx->Indirect = FALSE;
    ctx->Baud = 9600;
    ctx->Data.card = NULL;
    ctx->Data.t0 = NULL;
    ctx->Data.t1 = NULL;

    retry = 3;
    do  {
        rs232Drain(ctx->fh);
        rs232Flush(ctx->fh);
        rc = getFirmware(ctx);
    } while (rc && retry--);

    if (rc < 0) {
        free(ctx);
        return IFD_COMMUNICATION_ERROR;
    }
    
    pthread_mutex_lock(&ifdh_status_mutex);

    for (indx = 0; (indx < MAX_READER) && ecoTable[indx]; indx++);

    if (indx == MAX_READER) {
        pthread_mutex_unlock(&ifdh_status_mutex);
        return IFD_COMMUNICATION_ERROR;
    }
    
    ecoTable[indx] = ctx;
    
    pthread_mutex_unlock(&ifdh_status_mutex);
    
    return IFD_SUCCESS;
    
  
}


RESPONSECODE IFDHCloseChannel ( DWORD Lun ) {
  
  /* This function should close the reader communication channel
     for the particular reader.  Prior to closing the communication channel
     the reader should make sure the card is powered down and the terminal
     is also powered down.

     returns:

     IFD_SUCCESS
     IFD_COMMUNICATION_ERROR     
  */

    int rc;
    struct eco5000_t *ctx;
    struct memorycard_t *pt;
    
    rc = LookupLun(Lun);

    if (rc < 0) {
        return IFD_COMMUNICATION_ERROR;
    }
    
    ctx = ecoTable[rc];

    if (!ctx) {
        return IFD_COMMUNICATION_ERROR;
    }
    
    if (ctx->Data.card != NULL) {
      
        pt = ctx->Data.card;
      
        if (pt != NULL)
            free(pt);
    }

    rs232Mode(ctx->fh, 9600, 'E', 8, 1, 100);
    rs232Close(ctx->fh);
    free(ctx);
    ecoTable[rc] = NULL;
    
    return IFD_SUCCESS;

}


RESPONSECODE IFDHGetCapabilities ( DWORD Lun, DWORD Tag, 
				   PDWORD Length, PUCHAR Value ) {
  
  /* This function should get the slot/card capabilities for a particular
     slot/card specified by Lun.  Again, if you have only 1 card slot and don't mind
     loading a new driver for each reader then ignore Lun.

     Tag - the tag for the information requested
         example: TAG_IFD_ATR - return the Atr and it's size (required).
         these tags are defined in ifdhandler.h

     Length - the length of the returned data
     Value  - the value of the data

     returns:
     
     IFD_SUCCESS
     IFD_ERROR_TAG
  */


    int rc;
    struct eco5000_t *ctx;

    rc = LookupLun(Lun);

    if (rc < 0) {
        return IFD_COMMUNICATION_ERROR;
    }
    
    ctx = ecoTable[rc];

    if (!ctx) {
        return IFD_COMMUNICATION_ERROR;
    }
    
    switch(Tag) {

        case TAG_IFD_ATR:
            memcpy(Value, ctx->ATR, ctx->LenOfATR);
            *Length = ctx->LenOfATR;
            break;

        case TAG_IFD_SIMULTANEOUS_ACCESS:
            *Value = MAX_READER;
            *Length = 1;
            break;
            
        default:
            return IFD_ERROR_TAG;
            
    }
    
    return IFD_SUCCESS;    
}


RESPONSECODE IFDHSetCapabilities ( DWORD Lun, DWORD Tag, 
			       DWORD Length, PUCHAR Value ) {

  /* This function should set the slot/card capabilities for a particular
     slot/card specified by Lun.  Again, if you have only 1 card slot and don't mind
     loading a new driver for each reader then ignore Lun.

     Tag - the tag for the information needing set

     Length - the length of the returned data
     Value  - the value of the data

     returns:
     
     IFD_SUCCESS
     IFD_ERROR_TAG
     IFD_ERROR_SET_FAILURE
     IFD_ERROR_VALUE_READ_ONLY
  */

    return IFD_SUCCESS;
    
}

RESPONSECODE IFDHSetProtocolParameters ( DWORD Lun, DWORD Protocol, 
				   UCHAR Flags, UCHAR PTS1,
				   UCHAR PTS2, UCHAR PTS3) {

  /* This function should set the PTS of a particular card/slot using
     the three PTS parameters sent

     Protocol  - 0 .... 14  T=0 .... T=14
     Flags     - Logical OR of possible values:
     IFD_NEGOTIATE_PTS1 IFD_NEGOTIATE_PTS2 IFD_NEGOTIATE_PTS3
     to determine which PTS values to negotiate.
     PTS1,PTS2,PTS3 - PTS Values.

     returns:

     IFD_SUCCESS
     IFD_ERROR_PTS_FAILURE
     IFD_COMMUNICATION_ERROR
     IFD_PROTOCOL_NOT_SUPPORTED
  */

    int rc;
    struct eco5000_t *ctx;

    rc = LookupLun(Lun);

    if (rc < 0) {
        return IFD_COMMUNICATION_ERROR;
    }
    
    ctx = ecoTable[rc];

    if (!ctx) {
        return IFD_COMMUNICATION_ERROR;
    }
    
#ifdef DEBUG
    printf("[IFDHSetProtocolParameters] PPS request!!\n");
    printf("[IFDHSetProtocolParameters] Protocol = %u\n", Protocol);
    printf("[IFDHSetProtocolParameters] Flags = %u\n", Flags);
    printf("[IFDHSetProtocolParameters] PTS1 = %u\n", PTS1);
    printf("[IFDHSetProtocolParameters] PTS2 = %u\n", PTS2);
    printf("[IFDHSetProtocolParameters] PTS3 = %u\n", PTS3);
    printf("[IFDHSetProtocolParameters] Old protocol (from context) = %u\n", ctx->Protocol);
#endif

    rc = ecoCommand(ctx, RX_ON, 0, NULL, 0, NULL);
    
    if (rc < 0) {
       return IFD_COMMUNICATION_ERROR;
    }
    
    if (Flags == 0x00) {
        rc = PerformPPS(ctx, 0x10 | (Protocol - 1), (ctx->FI << 4) | ctx->DI, 0, 0);
    } else {
        rc = PerformPPS(ctx, Flags, PTS1, PTS2, PTS3); 
    }
    
    if (rc < 0) {
#ifdef DEBUG
        printf("[IFDHSetProtocolParameters] PPS failed! rc = %i\n", rc);
#endif
        return IFD_ERROR_PTS_FAILURE;        
    }

    if (ctx->Data.card != NULL) {
        free(ctx->Data.card);
    }
    
    rc = InstallHandler(ctx);

    if (rc < 0) {
#ifdef DEBUG
        printf("[IFDHSetProtocolParameters] PPS - cannot install handler! rc = %i\n", rc);
#endif
        return IFD_COMMUNICATION_ERROR;
    }

#ifdef DEBUG
    printf("[IFDHSetProtocolParameters] PPS success! New protocol = %u\n", ctx->Protocol);
#endif
    
    return IFD_SUCCESS;
    
}


RESPONSECODE IFDHPowerICC ( DWORD Lun, DWORD Action, 
			    PUCHAR Atr, PDWORD AtrLength ) {

  /* This function controls the power and reset signals of the smartcard reader
     at the particular reader/slot specified by Lun.

     Action - Action to be taken on the card.

     IFD_POWER_UP - Power and reset the card if not done so 
     (store the ATR and return it and it's length).
 
     IFD_POWER_DOWN - Power down the card if not done already 
     (Atr/AtrLength should
     be zero'd)
 
    IFD_RESET - Perform a quick reset on the card.  If the card is not powered
     power up the card.  (Store and return the Atr/Length)

     Atr - Answer to Reset of the card.  The driver is responsible for caching
     this value in case IFDHGetCapabilities is called requesting the ATR and it's
     length.  This should not exceed MAX_ATR_SIZE.

     AtrLength - Length of the Atr.  This should not exceed MAX_ATR_SIZE.

     Notes:

     Memory cards without an ATR should return IFD_SUCCESS on reset
     but the Atr should be zero'd and the length should be zero

     Reset errors should return zero for the AtrLength and return 
     IFD_ERROR_POWER_ACTION.

     returns:

     IFD_SUCCESS
     IFD_ERROR_POWER_ACTION
     IFD_COMMUNICATION_ERROR
     IFD_NOT_SUPPORTED
  */

    int rc;
    struct eco5000_t *ctx;
    unsigned char rsp[128];
    unsigned int lr;
    
    rc = LookupLun(Lun);

    if (rc < 0) {
        return IFD_COMMUNICATION_ERROR;
    }
    
    ctx = ecoTable[rc];

    if (!ctx) {
        return IFD_COMMUNICATION_ERROR;
    }
    
    lr = sizeof(rsp);
    
    switch (Action) {

        case IFD_RESET:

            
#ifdef DEBUG
            printf("[IFDHPowerICC] IFD_RESET called ...\n");
#endif

            rc = ResetTerminal(ctx, &lr, rsp);

            if (rc < 0) {    
#ifdef DEBUG
                printf("[IFDHPowerICC] IFD_RESET - communication error!\n");
#endif
                return IFD_COMMUNICATION_ERROR;
            }

            rc = Async_PowerOff(ctx);      
            
            if (rc < 0) {    
#ifdef DEBUG
                printf("[IFDHPowerICC] IFD_RESET - communication error!\n");
#endif
                return IFD_COMMUNICATION_ERROR;
            }
            
                /* make sure the ATR is not valid */
            memset(Atr, 0x00, sizeof(Atr));
            *AtrLength = 0;
            
            rc = IFDGetATR(ctx);

            if (rc < 0) {
#ifdef DEBUG
                printf("[IFDHPowerICC] IFD_RESET - communication error (IFDGetATR)! rc = %i\n", rc);
#endif
                return IFD_COMMUNICATION_ERROR;
            }

            rc = InstallHandler(ctx);

            if (rc < 0) {
#ifdef DEBUG
                printf("[IFDHPowerICC] IFD_RESET - communication error (InstallHandler)! rc = %i\n", rc);
#endif
                return IFD_COMMUNICATION_ERROR;
            }
            
                /* the ATR is valid - copy it */
            memcpy(Atr, ctx->ATR, ctx->LenOfATR);
            *AtrLength = ctx->LenOfATR;

            break;
            
        case IFD_POWER_UP:

#ifdef DEBUG
            printf("[IFDHPowerICC] IFD_POWER_UP called ...\n");
#endif

            rc = ResetTerminal(ctx, &lr, rsp);

            if (rc < 0) {
#ifdef DEBUG
                printf("[IFDHPowerICC] IFD_POWER_UP - communication error (ResetTerminal)! rc = %i\n", rc);
#endif
                return IFD_COMMUNICATION_ERROR;
            }

                /* make sure the ATR is not valid */
            memset(Atr, 0x00, sizeof(Atr));
            *AtrLength = 0;

            rc = IFDGetATR(ctx);
            
            if (rc < 0) {
#ifdef DEBUG
                printf("[IFDHPowerICC] IFD_POWER_UP - communication error (IFDGetATR)! rc = %i\n", rc);
#endif
                return IFD_COMMUNICATION_ERROR;
            }

            rc = InstallHandler(ctx);
            
            if (rc < 0) {
#ifdef DEBUG
                printf("[IFDHPowerICC] IFD_POWER_UP - communication error (InstallHandler)! rc = %i\n", rc);
#endif
                return IFD_COMMUNICATION_ERROR;
            }
            
                /* the ATR is valid - copy it */
            memcpy(Atr, ctx->ATR, ctx->LenOfATR);
            *AtrLength = ctx->LenOfATR;
            
            break;
            
        case IFD_POWER_DOWN:
            
#ifdef DEBUG
            printf("[IFDHPowerICC] IFD_POWER_DOWN called ...\n");
#endif

                /* We always power off async. card - this will also shutdown sync. cards. */ 
            rc = Async_PowerOff(ctx);
            
            if (rc < 0) {
                
#ifdef DEBUG
                printf("[IFDHPowerICC] IFD_POWER_DOWN - communication error!\n");
#endif
                return IFD_COMMUNICATION_ERROR;
            }
            
            memset(Atr, 0x00, sizeof(Atr));
            *AtrLength = 0;
            
            break;
            
        default:
#ifdef DEBUG
            printf("[IFDHPowerICC] IFD_NOT_SUPPORTED!\n");
#endif
            return IFD_NOT_SUPPORTED;
    }
    
    return IFD_SUCCESS;

}

RESPONSECODE IFDHTransmitToICC ( DWORD Lun, SCARD_IO_HEADER SendPci, 
				 PUCHAR TxBuffer, DWORD TxLength, 
				 PUCHAR RxBuffer, PDWORD RxLength, 
				 PSCARD_IO_HEADER RecvPci ) {
  
  /* This function performs an APDU exchange with the card/slot specified by
     Lun.  The driver is responsible for performing any protocol specific exchanges
     such as T=0/1 ... differences.  Calling this function will abstract all protocol
     differences.

     SendPci
     Protocol - 0, 1, .... 14
     Length   - Not used.

     TxBuffer - Transmit APDU example (0x00 0xA4 0x00 0x00 0x02 0x3F 0x00)
     TxLength - Length of this buffer.
     RxBuffer - Receive APDU example (0x61 0x14)
     RxLength - Length of the received APDU.  This function will be passed
     the size of the buffer of RxBuffer and this function is responsible for
     setting this to the length of the received APDU.  This should be ZERO
     on all errors.  The resource manager will take responsibility of zeroing
     out any temporary APDU buffers for security reasons.
  
     RecvPci
     Protocol - 0, 1, .... 14
     Length   - Not used.

     Notes:
     The driver is responsible for knowing what type of card it has.  If the current
     slot/card contains a memory card then this command should ignore the Protocol
     and use the MCT style commands for support for these style cards and transmit 
     them appropriately.  If your reader does not support memory cards or you don't
     want to then ignore this.

     RxLength should be set to zero on error.

     returns:
     
     IFD_SUCCESS
     IFD_COMMUNICATION_ERROR
     IFD_RESPONSE_TIMEOUT
     IFD_ICC_NOT_PRESENT
     IFD_PROTOCOL_NOT_SUPPORTED
  */


    int rc;
    struct eco5000_t *ctx;
    unsigned char status;
    int llr;
    int llc;
    

    rc = LookupLun(Lun);

    if (rc < 0) {
#ifdef DEBUG
        printf("[IFDHTransmitToICC] LUN not found!\n");
#endif
        return IFD_COMMUNICATION_ERROR;
    }
    
    ctx = ecoTable[rc];

    if (!ctx) {
#ifdef DEBUG
        printf("[IFDHTransmitToICC] ctx not found!\n");
#endif    
        return IFD_COMMUNICATION_ERROR;
    }

    rc = IFDHICCPresence(Lun);

    if (rc != IFD_ICC_PRESENT) {
#ifdef DEBUG
        printf("[IFDHTransmitToICC] Error: no icc present!\n");
#endif
        return IFD_ICC_NOT_PRESENT;
    }


        /* We can only check the active protocol when we us processor cards -
           otherwise the content of the SendPci-structure is not valid for us. */

    if (ctx->Type == ASYNC_ICC) {
        
        if (ctx->Protocol != SendPci.Protocol) {
#ifdef DEBUG
            printf("[IFDHTransmitToICC] Protocol mismatch: %u (card) vs. %u (func)!\n", ctx->Protocol, SendPci.Protocol);
#endif
            return IFD_COMMUNICATION_ERROR;
        }
    }
    
    llc = (int) TxLength;
    llr = (int) *RxLength;
    

    if (!(*ctx->CTModFunc)) {
#ifdef DEBUG
        printf("[IFDHTransmitToICC] Error in CTModFunc(). No valid context!\n");
#endif
        return IFD_COMMUNICATION_ERROR;
    }
    

    rc = (*ctx->CTModFunc)(ctx, llc, TxBuffer, &llr, RxBuffer);


    if (rc < 0) {
#ifdef DEBUG
        printf("[IFDHTransmitToICC] Error in protocol function. rc = %i !\n", rc);
#endif
        return IFD_COMMUNICATION_ERROR;
    }

    *RxLength = llr;
    
    return IFD_SUCCESS;
    
}

RESPONSECODE IFDHControl ( DWORD Lun, PUCHAR TxBuffer, 
			 DWORD TxLength, PUCHAR RxBuffer, 
			 PDWORD RxLength ) {

  /* This function performs a data exchange with the reader (not the card)
     specified by Lun.  Here XXXX will only be used.
     It is responsible for abstracting functionality such as PIN pads,
     biometrics, LCD panels, etc.  You should follow the MCT, CTBCS 
     specifications for a list of accepted commands to implement.

     TxBuffer - Transmit data
     TxLength - Length of this buffer.
     RxBuffer - Receive data
     RxLength - Length of the received data.  This function will be passed
     the length of the buffer RxBuffer and it must set this to the length
     of the received data.

     Notes:
     RxLength should be zero on error.
  */

    return IFD_NOT_SUPPORTED;
    
}

RESPONSECODE IFDHICCPresence( DWORD Lun ) {

  /* This function returns the status of the card inserted in the 
     reader/slot specified by Lun.  It will return either:

     returns:
     IFD_ICC_PRESENT
     IFD_ICC_NOT_PRESENT
     IFD_COMMUNICATION_ERROR
  */

    int rc;
    struct eco5000_t *ctx;
    unsigned char status;


    rc = LookupLun(Lun);
    
    if (rc < 0) {
#ifdef DEBUG
        printf("[IFDHICCPresence] LUN not found!\n");
#endif      
        return IFD_COMMUNICATION_ERROR;
    }
    
    ctx = ecoTable[rc];

    if (!ctx) {
#ifdef DEBUG
        printf("[IFDHICCPresence] ctx not found!\n");
#endif    
        return IFD_COMMUNICATION_ERROR;
    }
    
    if ((rc = ecoCommand(ctx, READ_CARD_IN, 0, NULL, 1, &status)) < 0) {
#ifdef DEBUG
        printf("[IFDHICCPresence] ecoCommand error! rc = %i\n", rc);
#endif
        return IFD_COMMUNICATION_ERROR;
    }

    if (status == 0x00)   {
        rc = IFD_ICC_NOT_PRESENT;
    } else {
        rc = IFD_ICC_PRESENT;
    }
    
    return rc;
    
}
